package edu.dhbw.mannheim.tigers.sim.util.network;


import java.io.EOFException;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.util.collection.Pair;


/**
 * This class should be able to register to a multicast group, and receive {@link DatagramPacket}s
 * on it
 * 
 * @author Gero
 * 
 */
public class MulticastUDPReceiver implements IReceiver {
	
	protected final Logger												log				= Logger.getLogger(getClass());
	

	// Connection
	private MulticastSocket												socket			= null;
	private InetAddress													group				= null;
	
	private List<Pair<InetSocketAddress, NetworkInterface>>	groups			= null;
	
	
	/**
	  * @param port
	  * @param group
	  */
	public MulticastUDPReceiver(int port, String group) {
		this(port, group, (NetworkInterface) null);
	}
	
	
	/**
	  * @param port
	  * @param groupStr
	  * @param nif
	  */
	public MulticastUDPReceiver(int port, String groupStr, NetworkInterface nif) {
		this(port, nif);
		
		joinNetworkGroup(groupStr);
	}
	
	
	private void joinNetworkGroup(String groupStr) {
		
		// Parse group
		try {
			group = InetAddress.getByName(groupStr);
		} catch (UnknownHostException err) {
			log.error("Unable to read multicast group address!", err);
		}
		
		// Join group
		try {
			socket.joinGroup(group);
		} catch (IOException err) {
			log.error("Could not resolve address: " + group, err);
			return;
		}
	}
	
	
	
	/**
	  * @see MulticastUDPReceiver
	  * @param port
	  * @param pairs
	  */
	public MulticastUDPReceiver(int port, List<Pair<InetSocketAddress, NetworkInterface>> pairs) {
		this(port, (NetworkInterface) null);
		
		joinGroups(pairs);
	}
	
	
	/**
	  * @see MulticastUDPReceiver
	  * @param port
	  * @param nif
	  */
	private MulticastUDPReceiver(int port, NetworkInterface nif) {
		
		// Create socket
		boolean socketError = true;
		while (socket == null && socketError) { // Tries to open an socket until it works! =)
			try {
				socket = new MulticastSocket(port);
				
				// Set nif
				if (nif != null) {
					socket.setNetworkInterface(nif);
				}
			} catch (SocketException err) {
				log.warn("Error while opening a socket on port: '" + port + "'! (UDP; maybe port blocked???)", err);
				port++;
				log.debug("Trying to open a socket on port '" + port + "'...");
			} catch (IOException err) {
				socketError = false;
				log.error("Error while opening a socket! (I/O)!", err);
			}
		}
	}
	

	/**
	 * @see MulticastUDPReceiver
	 * @param socket
	 */
	MulticastUDPReceiver(MulticastSocket socket) {
		this.socket = socket;
	}
	

	/**
	 * @see MulticastUDPReceiver
	 * @param socket
	 * @param groupStr
	 * @param nif
	 */
	MulticastUDPReceiver(MulticastSocket socket, String groupStr, NetworkInterface nif) {
		this.socket = socket;
		
		if (nif != null) {
			try {
				socket.setNetworkInterface(nif);
			} catch (SocketException err) {
				log.error("Unable to set nif: " + nif);
			}
		}
		
		joinNetworkGroup(groupStr);
	}
	
	
	/**
	  * @see MulticastUDPReceiver
	  * @param socket
	  * @param pairs
	  */
	MulticastUDPReceiver(MulticastSocket socket, List<Pair<InetSocketAddress, NetworkInterface>> pairs) {
		this.socket = socket;
		
		joinGroups(pairs);
	}
	
	
	private void joinGroups(List<Pair<InetSocketAddress, NetworkInterface>> pairs) {
		groups = new ArrayList<Pair<InetSocketAddress, NetworkInterface>>();
		
		for (Pair<InetSocketAddress, NetworkInterface> pair : pairs) {
			InetSocketAddress group = pair.left;
			if (group.getAddress().isMulticastAddress()) {
				NetworkInterface iface = pair.right;
				
				try {
					socket.joinGroup(group, iface);
					groups.add(pair);
				} catch (IOException err) {
					log.debug("Error joining group '" + group + "' on nif '" + iface + "'.");
					continue;
				}
			}
		}
	}
	

	@Override
	public DatagramPacket receive(DatagramPacket store) throws IOException {
		try {
			socket.receive(store); // New data is written into the packets buffer
		} catch (EOFException eof) {
			log.error("EOF error, buffer may be too small!", eof);
		}
		
		return store;
	}
	

	@Override
	public void cleanup() throws IOException {
		if (socket != null && !socket.isClosed()) {
			
			if (groups == null)
			{
				try {
					if (group != null) {
						socket.leaveGroup(group);
					}
				} catch (IOException err) {
					log.error("Error while leaving multicast group '" + group + "'!", err);
				}
				
				group = null;
			} else {
				
				for (Pair<InetSocketAddress, NetworkInterface> pair : groups) {
					InetSocketAddress group = pair.left;
					NetworkInterface nif = pair.right;
					
					try {
						socket.leaveGroup(group, nif);
					} catch (IOException err) {
						log.error("Error while leaving multicast group '" + group + "' on nif '" + nif + "'!", err);
					}
				}
				
				groups = null;
			}
		}
		
		if (socket != null) {
			socket.close();
			socket = null;
		}
	}
	

	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public InetAddress getLocalAddress() {
		return socket.getLocalAddress();
	}
	

	@Override
	public int getLocalPort() {
		return socket.getLocalPort();
	}
	

	DatagramSocket getSocket() {
		return socket;
	}
}
